import logging

from .links import Links
from .sentences_detailed import SentencesDetailed
from .tatoebatools import Tatoeba

logger = logging.getLogger(__name__)


class ParallelCorpus:
    """A parallel corpus of bilingual sentences' pairs

    A parallel corpus between two languages is a collection of the
    Tatoeba sentences in the source language placed alongside their
    translations in the target language.
    """

    def __init__(
        self,
        source_language_code,
        target_language_code,
        update=True,
        verbose=True,
    ):
        """
        Parameters
        ----------
        source_language_code : str
            The ISO 639-3 code of the parallel corpus' source language
        target_language_code : str
             The ISO 639-3 code of the parallel corpus' target language
        update : bool, optional
            Whether an update of the local data is forced or not, by
            default True
        verbose : bool, optional
            The verbosity of the logging, by default True
        """

        self._src_lg = source_language_code
        self._tgt_lg = target_language_code
        self._upd = update
        self._vb = verbose

        # the source sentences
        self._sentences = {}
        # the target sentences
        self._translations = {}

        self._update()  # update local data
        self._load()  # load the sentences' data in memory

    def __iter__(self):
        """
        Yields
        -------
        tuple
            the SentenceDetailed instances of both the source sentence
            and its translation
        """

        if self._sentences and self._translations:
            for lk in Links(self._src_lg, self._tgt_lg):
                if (
                    lk.sentence_id in self._sentences
                    and lk.translation_id in self._translations
                ):
                    yield (
                        self._sentences[lk.sentence_id],
                        self._translations[lk.translation_id],
                    )

    def _update(self):
        """Updates local data required for this parallel corpus"""
        tatoeba = Tatoeba()
        if self._upd:  # if update explicitly demanded
            tatoeba.update(
                ["sentences_detailed", "links"],
                [self._src_lg, self._tgt_lg],
                oriented_pair=True,
                verbose=False,
            )
        else:  # if necessary local data missing
            if not SentencesDetailed(self._src_lg).version:
                tatoeba.update(
                    ["sentences_detailed"],
                    [self._src_lg],
                    verbose=False,
                )
            if not SentencesDetailed(self._tgt_lg).version:
                tatoeba.update(
                    ["sentences_detailed"],
                    [self._tgt_lg],
                    verbose=False,
                )
            if not Links(self._src_lg, self._tgt_lg).version:
                tatoeba.update(
                    ["links"],
                    [self._src_lg, self._tgt_lg],
                    oriented_pair=True,
                    verbose=False,
                )

    def _load(self):
        """Loads the sentences and links from their datafiles"""

        links = Links(self._src_lg, self._tgt_lg)
        src_corpus = SentencesDetailed(self._src_lg)
        tgt_corpus = SentencesDetailed(self._tgt_lg)

        if any(not tbl.version for tbl in (links, src_corpus, tgt_corpus)):
            logger.error(
                f"{self._src_lg}-{self._tgt_lg} parralel corpus: "
                f"missing data file(s). please update your data."
            )
        elif not (
            links.version.date()
            == src_corpus.version.date()
            == tgt_corpus.version.date()
        ):
            logger.error(
                f"{self._src_lg}-{self._tgt_lg} parralel corpus: "
                f"data files' versions difer. please update your data."
            )
        else:
            src_ids, tgt_ids = links.ids
            self._sentences = src_corpus.get(src_ids)
            self._translations = tgt_corpus.get(tgt_ids)



import logging
from datetime import datetime, timedelta

import requests
from bs4 import BeautifulSoup

from .config import DATA_DIR

logger = logging.getLogger(__name__)


class DownloadPages:
    """Web pages at downloads.tatoeba.org from
    which export files' versions can be scraped
    """

    dir = DATA_DIR.joinpath("export_pages")
    dir.mkdir(parents=True, exist_ok=True)

    def get_versions(self, url):
        """Scraps the versions of the files listed in the web page

        Returns
        -------
        dict
            the versions of all urls listed in the web page
        """
        html = self._get_html(url)
        versions = _extract_versions(html)

        return {self.url + k: v for k, v in versions.items()}

    def get_names(self, url):
        """Scraps the directory names listed in the web page

        Returns
        -------
        list
            the names of all urls listed in the web page
        """
        html = self._get_html(url)

        return _extract_names(html)

    def _get_html(self, url, update_time=5):
        """Get the HTML content of this web page

        Parameters
        ----------
        update_time : int, optional
            the time in minutes after which the local file of the web page
            is re-downloaded, by default 5
        """
        self.url = url if url.endswith("/") else f"{url}/"

        if not self.version or self.version < datetime.now() - timedelta(
            minutes=update_time
        ):
            # download a newer version
            try:
                r = requests.get(self.url)
                with open(self.path, "w") as f:
                    f.write(r.text)
            except requests.exceptions.RequestException:
                logger.warning(f"error while requesting {self.url}")
                return ""
            else:
                return r.text
        else:
            # load up to date local file
            with open(self.path) as f:
                return f.read()

    @property
    def version(self):
        """Get the version of the local web page

        Returns
        -------
        datetime
            the last modification time of the file if this one exists
        """
        if self.path.is_file():
            return datetime.fromtimestamp(self.path.stat().st_mtime)
        else:
            return

    @property
    def path(self):
        """Get the local path of the web page

        Returns
        -------
        Path
            the local path of the file
        """
        stem = self.url[:-1].rsplit("/", 1)[-1]

        return self.dir.joinpath(f"{stem}.html")


def _extract_versions(html):
    """Extracts the versions of the files from an export page HTML code"""
    soup = BeautifulSoup(html, features="html.parser").find("pre")
    texts = [x.strip() for x in soup.findAll(text=True) if soup]

    return {
        x: datetime.strptime(texts[i + 1][:17], "%d-%b-%Y %H:%M")
        for i, x in enumerate(texts)
        if i % 2 == 0 and texts[i + 1] and x[-1] != "/"
    }


def _extract_names(html):
    """Extracts the names of the directories from an export page HTML code"""
    soup = BeautifulSoup(html, features="html.parser")
    links = [a.get("href") for a in soup.find_all("a") if soup]

    return [lk[:-1] for lk in links if lk[:-1].isalpha()]



from io import StringIO
from unittest.mock import patch

from tatoebatools.datafile import (
    DataFile,
    _custom_reader,
    _get_mapped_fields,
    _unsplit_field,
)
from tatoebatools.exceptions import NoDataFile


class TestUnsplitField:
    def test_without_split_field(self):
        row = ["foobar", "text not split by delimiters"]
        new_row = _unsplit_field(row, nb_cols=2, delimiter=",", index_field=-1)
        assert new_row == row

    def test_with_end_split_field(self):
        row = ["foobar", "text", " split", " by", " delimiters"]
        ok_row = ["foobar", "text, split, by, delimiters"]
        new_row = _unsplit_field(row, nb_cols=2, delimiter=",", index_field=-1)
        assert new_row == ok_row

    def test_with_midlle_split_field(self):
        row = ["foobar", "text", " split", " by", " delimiters", "foo"]
        ok_row = ["foobar", "text, split, by, delimiters", "foo"]
        new_row = _unsplit_field(row, nb_cols=3, delimiter=",", index_field=1)
        assert new_row == ok_row


class TestCustomReader:
    def test_with_commas(self):
        str_io = StringIO("a,b,c\n")
        rows = [row for row in _custom_reader(str_io, ",", -1)]
        assert rows == [["a", "b", "c"]]

    def test_with_multiple_rows(self):
        str_io = StringIO("a,b,c\nd,e,f\n")
        rows = [row for row in _custom_reader(str_io, ",", -1)]
        assert rows == [["a", "b", "c"], ["d", "e", "f"]]

    def test_with_tabs(self):
        str_io = StringIO("a\tb\tc\n")
        rows = [row for row in _custom_reader(str_io, "\t", -1)]
        assert rows == [["a", "b", "c"]]

    def test_with_null_field(self):
        str_io = StringIO("a,\\N,c\n")
        rows = [row for row in _custom_reader(str_io, ",", -1)]
        assert rows == [["a", "\\N", "c"]]

    def test_with_split_end_column(self):
        str_io = StringIO("a,b,c\nd,e,f,f")
        rows = [row for row in _custom_reader(str_io, ",", -1)]
        assert rows == [["a", "b", "c"], ["d", "e", "f,f"]]

    def test_with_split_middle_column(self):
        str_io = StringIO("a,b,c\nd,e,e,f")
        rows = [row for row in _custom_reader(str_io, ",", 1)]
        assert rows == [["a", "b", "c"], ["d", "e,e", "f"]]

    def test_with_multiline_row(self):
        str_io = StringIO(
            "abk\t4\tbackstreetboys\tunited states\n"
            "\\\n"
            "American flag\n"
            "\\\n"
            "\n"
        )
        rows = [row for row in _custom_reader(str_io, "\t", -1)]
        assert rows == [
            [
                "abk",
                "4",
                "backstreetboys",
                "united states \\ American flag \\ ",
            ]
        ]


class TestGetMappedFields:
    def test_without_index(self):
        row = ["foo", "bar"]
        columns = [1]
        index = {}
        assert _get_mapped_fields(row, columns, index) == ["bar"]

    def test_with_too_short_row(self):
        row = ["foo", "bar"]
        columns = [2]
        index = {}
        assert _get_mapped_fields(row, columns, index) == [""]

    def test_with_index(self):
        row = ["23", "42"]
        index = {"23": "foo", "42": "bar"}
        columns = [0, 1]
        assert _get_mapped_fields(row, columns, index) == ["foo", "bar"]


class TestDataFile:
    fake_string_io = StringIO("42,fra,foo\n123,eng,bar\n")
    fake_table = [
        ["42", "fra", "foo"],
        ["123", "eng", "bar"],
    ]

    @patch("builtins.open")
    @patch("tatoebatools.datafile.Path.is_file")
    def test_iter_with_file(self, m_isfile, m_open):
        m_isfile.return_value = True
        m_open.return_value = self.fake_string_io
        df = DataFile("any_file_path", delimiter=",", text_col=-1)

        assert [row for row in df] == self.fake_table
        assert m_isfile.call_count == 1
        assert m_open.call_count == 1

    @patch("tatoebatools.datafile.DataFile.__iter__")
    def test_index_with_ok_rows(self, m_iter):
        m_iter.return_value = iter(self.fake_table)
        df = DataFile("any_file_path", delimiter=",", text_col=-1)

        assert df.index(0, 1) == {"42": "fra", "123": "eng"}
        assert m_iter.call_count == 1

    @patch("tatoebatools.datafile.DataFile.__iter__")
    def test_index_with_bad_rows(self, m_iter):
        m_iter.return_value = iter([["42", "fra", "foobar"], ["123"]])
        df = DataFile("any_file_path", delimiter=",", text_col=-1)

        assert df.index(0, 1) == {"42": "fra"}
        assert m_iter.call_count == 1

    @patch("tatoebatools.datafile.DataFile.__iter__")
    def test_index_with_no_datafile(self, m_iter):
        m_iter.side_effect = NoDataFile
        df = DataFile("any_file_path", delimiter=",", text_col=-1)

        assert df.index(0, 1) == {}
        assert m_iter.call_count == 1

    @patch("tatoebatools.datafile.Buffer")
    @patch("tatoebatools.datafile.tqdm")
    @patch("tatoebatools.datafile.logger")
    @patch("tatoebatools.datafile.DataFile.__iter__")
    def test_split_without_index(self, m_iter, m_logger, m_tqdm, m_buffer):
        m_iter.return_value = iter(self.fake_table)
        fp = "any_file_path"
        df = DataFile(fp, delimiter=",", text_col=-1)
        df.split(columns=[1])

        add_args = m_buffer.return_value.add.call_args_list
        assert add_args[0][0][0] == self.fake_table[0]
        assert add_args[1][0][0] == self.fake_table[1]
        assert add_args[0][0][1] != add_args[1][0][1]
        assert m_iter.call_count == 1
        assert m_buffer.return_value.clear.call_count == 1
        assert m_logger.info.call_count == 1
        assert m_tqdm.return_value.update.call_count == 2
        assert m_tqdm.return_value.close.call_count == 1

    @patch("tatoebatools.datafile.Buffer")
    @patch("tatoebatools.datafile.tqdm")
    @patch("tatoebatools.datafile.logger")
    @patch("tatoebatools.datafile.DataFile.__iter__")
    def test_split_with_index(self, m_iter, m_logger, m_tqdm, m_buffer):
        m_iter.return_value = iter(self.fake_table)
        fp = "any_file_path"
        df = DataFile(fp, delimiter=",", text_col=-1)
        ind = {"42": "fra", "123": "eng"}
        df.split(columns=[0], index=ind)

        add_args = m_buffer.return_value.add.call_args_list
        assert add_args[0][0][0] == self.fake_table[0]
        assert add_args[1][0][0] == self.fake_table[1]
        assert add_args[0][0][1] != add_args[1][0][1]
        assert m_iter.call_count == 1
        assert m_buffer.return_value.clear.call_count == 1
        assert m_logger.info.call_count == 1
        assert m_tqdm.return_value.update.call_count == 2
        assert m_tqdm.return_value.close.call_count == 1

    def test_get_out_filename_csv(self):
        fp = "any_file_path"
        df = DataFile(fp, delimiter=",", text_col=-1)

        assert df._get_out_filename(["eng", "fra"]) == f"eng-fra_{fp}.csv"

    def test_get_out_filename_tsv(self):
        fp = "any_file_path"
        df = DataFile(fp, delimiter="\t", text_col=-1)

        assert df._get_out_filename(["eng", "fra"]) == f"eng-fra_{fp}.tsv"



from tatoebatools.download_page import _extract_versions
from datetime import datetime


def test_extract_versions():

    html = (
        "<html>\r\n"
        "<head><title>Index of /exports/per_language/eng/</title></head>\r\n"
        '<body bgcolor="white">\r\n'
        "<h1>Index of /exports/per_language/eng/</h1>"
        "<hr><pre>"
        '<a href="../">../</a>\r\n'
        '<a href="eng_sentences.tsv.bz2">'
        "eng_sentences.tsv.bz2</a>"
        "                              23-May-2020 06:25          14828881\r\n"
        '<a href="eng_sentences_CC0.tsv.bz2">'
        "eng_sentences_CC0.tsv.bz2</a>"
        "                          23-May-2020 06:25             96899\r\n"
        '<a href="eng_sentences_detailed.tsv.bz2">'
        "eng_sentences_detailed.tsv.bz2</a>"
        "                     23-May-2020 06:23          19841380\r\n"
        "</pre><hr>"
        "</body>\r\n"
        "</html>\r\n"
    )

    ok_versions = {
        "eng_sentences.tsv.bz2": datetime(2020, 5, 23, 6, 25),
        "eng_sentences_CC0.tsv.bz2": datetime(2020, 5, 23, 6, 25),
        "eng_sentences_detailed.tsv.bz2": datetime(2020, 5, 23, 6, 23),
    }

    assert _extract_versions(html) == ok_versions



import logging

from .config import DATA_DIR
from .datafile import DataFile
from .exceptions import NoDataFile
from .utils import get_extended_name, lazy_property
from .version import version

logger = logging.getLogger(__name__)


class SentencesBase:
    """The Tatoeba file containing the sentences' bases in a given
    language
    """

    _table = "sentences_base"
    _dir = DATA_DIR.joinpath(_table)

    def __init__(self, language, scope="all"):

        # the language code of the sentences (ISO-639 code most of the time)
        self._lg = language
        # rows that are iterated through
        self._sp = scope

    def __iter__(self):

        if self._sp == "all":
            fpath = self.path
        else:
            fname = get_extended_name(self.path, self._sp)
            fpath = SentencesBase._dir.joinpath(fname)

        try:
            fieldnames = [
                "sentence_id",
                "base_of_the_sentence",
            ]

            for row in DataFile(fpath, delimiter="\t"):
                row = {fieldnames[i]: x for i, x in enumerate(row)}
                yield SentenceBase(**row)
        except NoDataFile:
            msg = (
                f"no '{self._sp}' data locally available for the "
                f"'{SentencesBase._table}' table in {self._lg}."
            )

            logger.warning(msg)

    @property
    def language(self):
        """Get the language of the sentences"""
        return self._lg

    @property
    def stem(self):
        """Get the stem of the file of these sentences"""
        return f"{self._lg}_{SentencesBase._table}"

    @property
    def filename(self):
        """Get the name of the file of these sentences"""
        return f"{self.stem}.tsv"

    @property
    def path(self):
        """Get the path of the sentences' datafile"""
        return SentencesBase._dir.joinpath(self.filename)

    @lazy_property
    def version(self):
        """Get the version of the downloaded data of these sentences."""
        return version[self.stem]


class SentenceBase:
    """The base of a sentence from the Tatoeba corpus

    A sentence is based on another if it has been initially added as a
    translation.

    The base of a sentence is equal to:
    - 0 when the sentence is original (i.e. not based on another)
    - the sentence id it is based upon when this one is available
    - None when the information about the base is not available
    """

    def __init__(
        self,
        sentence_id,
        base_of_the_sentence,
    ):

        self._id = sentence_id
        self._bs = base_of_the_sentence

    @property
    def sentence_id(self):
        """Get the id of the sentence"""
        return int(self._id)

    @property
    def base_of_the_sentence(self):
        """Get the base of the sentence"""
        return int(self._bs) if self._bs != "\\N" else None



import logging

from .config import DATA_DIR
from .datafile import DataFile
from .version import version

logger = logging.getLogger(__name__)


class Table:
    """A Tatoeba table."""

    def __init__(self, name, languages):
        # the name of this table
        self._name = name
        # the languages in which some data is required
        self._lgs = languages

    def index(self, key_column, value_column):
        """Get the index that maps 2 columns of this table."""
        tbl_index = {}
        for df in self.language_detafiles:
            tbl_index.update(df.index(key_column, value_column))

        return tbl_index

    def classify(self, language_index=None):
        """Classify this table data by language."""
        if self.name == "user_languages":
            self.main_datafile.split(columns=[0])
        elif self.name == "queries":
            self.main_datafile.split(columns=[1])
        elif self.name == "links" and language_index:
            self.main_datafile.split(columns=[0, 1], index=language_index)
        elif self.name == "tags" and language_index:
            self.main_datafile.split(columns=[0], index=language_index)
        elif self.name == "sentences_in_lists" and language_index:
            self.main_datafile.split(columns=[1], index=language_index)
        elif self.name == "jpn_indices" and language_index:
            self.main_datafile.split(columns=[0], index=language_index)
        elif self.name == "sentences_with_audio" and language_index:
            self.main_datafile.split(columns=[0], index=language_index)
        else:
            return []

        for fs in self.language_filestems:
            version[fs] = self.main_datafile.version

        return self.language_detafiles

    @property
    def name(self):
        """Get the name of this table."""
        return self._name

    @property
    def path(self):
        """Get the local path of this table."""
        return DATA_DIR.joinpath(self._name)

    @property
    def main_datafile(self):
        """Get the multilingual datafile of this table."""
        main_filename = f"{self.name}.csv"
        delimiter = "," if self.name == "queries" else "\t"
        fp = self.path.joinpath(main_filename)

        return DataFile(fp, delimiter=delimiter)

    @property
    def language_detafiles(self):
        """Get the monolingual datafiles of this table."""
        language_datafiles = []
        for fn in self.language_filenames:
            fp = self.path.joinpath(fn)
            delimiter = "\t" if fp.suffix == ".tsv" else ","
            language_datafiles.append(DataFile(fp, delimiter=delimiter))

        return language_datafiles

    @property
    def language_filenames(self):
        """Get the names of the monolingual datafiles"""
        if self.name in (
            "sentences_detailed",
            "sentences_CC0",
            "transcriptions",
            "tags",
            "sentences_in_lists",
            "jpn_indices",
            "sentences_with_audio",
            "user_languages",
        ):
            return [f"{lg}_{self.name}.tsv" for lg in self._lgs]
        elif self.name in ("queries",):
            return [f"{lg}_{self.name}.csv" for lg in self._lgs]
        elif self.name in ("links",):
            return [
                f"{lg1}-{lg2}_{self.name}.tsv"
                for lg1 in self._lgs
                for lg2 in self._lgs
            ]
        else:
            return []

    @property
    def language_filestems(self):
        """Get the stems of the monolingual datafiles"""
        return [fn.rsplit(".")[0] for fn in self.language_filenames]



import logging

from .config import DATA_DIR
from .datafile import DataFile
from .exceptions import NoDataFile
from .utils import get_extended_name, lazy_property
from .version import version

logger = logging.getLogger(__name__)


class SentencesWithAudio:
    """The Tatoeba sentences with audio for a given language."""

    _table = "sentences_with_audio"
    _dir = DATA_DIR.joinpath(_table)

    def __init__(self, language, scope="all"):
        # the language of the sentences with audio
        self._lg = language
        # rows that are iterated through
        self._sp = scope

    def __iter__(self):

        if self._sp == "all":
            fpath = self.path
        else:
            fname = get_extended_name(self.path, self._sp)
            fpath = SentencesWithAudio._dir.joinpath(fname)

        try:
            fieldnames = [
                "sentence_id",
                "username",
                "license",
                "attribution_url",
            ]

            for row in DataFile(fpath, delimiter="\t"):
                row = {fieldnames[i]: x for i, x in enumerate(row)}

                yield SentenceWithAudio(**row)

        except NoDataFile:
            msg = (
                f"no '{self._sp}'data locally available for the '{self.table}'"
                f" table in {self._lg}."
            )

            logger.warning(msg)

    @property
    def language(self):
        """Get the language of these sentences with audio."""
        return self._lg

    @property
    def path(self):
        """Get the path where the sentences with audio are saved."""
        return SentencesWithAudio._dir.joinpath(self.filename)

    @property
    def stem(self):
        """Get the stem of the name of the file where the sentences
        with audio are saved
        """
        return f"{self._lg}_{SentencesWithAudio._table}"

    @property
    def filename(self):
        """Get the name of the file where the sentences with audio
        are saved
        """
        return f"{self.stem}.tsv"

    @lazy_property
    def version(self):
        """Get the version of these sentences with audio"""
        return version[self.stem]


class SentenceWithAudio:
    """A Tatoeba sentence with audio."""

    def __init__(self, sentence_id, username, license, attribution_url):

        self._id = sentence_id
        self._usr = username
        self._lic = license
        self._atr = attribution_url

    @property
    def sentence_id(self):
        """The id of the sentence with audio."""
        return int(self._id)

    @property
    def username(self):
        """The username that added the sentence with audio."""
        return self._usr

    @property
    def license(self):
        """The license of the sentence with audio."""
        return self._lic if self._lic != "\\N" else ""

    @property
    def attribution_url(self):
        """The url to the attrbution of the sentence with audio."""
        return self._atr if self._atr != "\\N" else ""



import logging
from datetime import datetime

from .config import DATA_DIR
from .datafile import DataFile
from .exceptions import NoDataFile
from .utils import get_extended_name, lazy_property
from .version import version

logger = logging.getLogger(__name__)


class UserLists:
    """The lists of sentences that the users have built."""

    _table = "user_lists"
    _dir = DATA_DIR.joinpath(_table)
    _filename = f"{_table}.csv"
    _path = _dir.joinpath(_filename)

    def __init__(self, scope="all"):

        self._sp = scope

    def __iter__(self):

        if self._sp == "all":
            fpath = self.path
        else:
            fname = get_extended_name(self.path, self._sp)
            fpath = UserLists._dir.joinpath(fname)

        try:
            fieldnames = [
                "list_id",
                "username",
                "date_created",
                "date_last_modified",
                "list_name",
                "editable_by",
            ]

            for row in DataFile(fpath, delimiter="\t", text_col=4):
                row = {fieldnames[i]: x for i, x in enumerate(row)}

                yield UserList(**row)

        except NoDataFile:
            msg = (
                f"no '{self._sp}' data locally available for the "
                f"'{UserLists._table}' table."
            )

            logger.warning(msg)

    @property
    def filename(self):
        """Get the name of the file of these tagged sentences."""
        return UserLists._filename

    @property
    def path(self):
        """Get the path of the tagged sentences' datafile."""
        return UserLists._path

    @lazy_property
    def version(self):
        """Get the version of the downloaded data of these tagged sentences."""
        return version[UserLists._table]


class UserList:
    """A list of sentences built by a Tatoeba user."""

    def __init__(
        self,
        list_id,
        username,
        date_created,
        date_last_modified,
        list_name,
        editable_by,
    ):

        self._id = list_id
        self._usr = username
        self._dcr = date_created
        self._dlm = date_last_modified
        self._nm = list_name
        self._edb = editable_by

    @property
    def list_id(self):
        """Get the id of this list."""
        return int(self._id)

    @property
    def username(self):
        """Get the name of the user that built this list."""
        return self._usr

    @property
    def date_created(self):
        """Get the date when this list has been created."""
        try:
            dt = datetime.strptime(self._dcr, "%Y-%m-%d %H:%M:%S")
        except ValueError:
            dt = None
        finally:
            return dt

    @property
    def date_last_modified(self):
        """Get the date when this list has been modified for the last time."""
        try:
            dt = datetime.strptime(self._dlm, "%Y-%m-%d %H:%M:%S")
        except ValueError:
            dt = None
        finally:
            return dt

    @property
    def list_name(self):
        """Get the name of this list."""
        return self._nm

    @property
    def editable_by(self):
        """Get the users that can edit this list."""
        return self._edb



from pathlib import Path

from pkg_resources import resource_filename

DATA_DIR = Path(resource_filename(__package__, "data"))

SUPPORTED_TABLES = (
    "sentences_base",
    "sentences_detailed",
    "sentences_CC0",
    "transcriptions",
    "links",
    "tags",
    "user_lists",
    "sentences_in_lists",
    "jpn_indices",
    "sentences_with_audio",
    "user_languages",
    "queries",
)

INDEX_SPLIT_TABLES = ()

SIMPLE_SPLIT_TABLES = ("queries",)

DIFFERENCE_TABLES = (
    "sentences_base",
    "sentences_detailed",
    "sentences_CC0",
    "transcriptions",
    "links",
    "tags",
    "user_lists",
    "sentences_in_lists",
    "jpn_indices",
    "sentences_with_audio",
    "user_languages",
)



import bz2
import csv
import logging
import tarfile
from pathlib import Path

import requests
from tqdm import tqdm

logger = logging.getLogger(__name__)


def download(from_url, to_directory):
    """Download a file. Overwrite previous version."""
    # build out file path
    filename = from_url.rsplit("/", 1)[-1]
    to_dir_path = Path(to_directory)
    to_dir_path.mkdir(parents=True, exist_ok=True)
    to_path = to_dir_path.joinpath(filename)

    try:
        with requests.get(from_url, stream=True) as r:
            r.raise_for_status()
            # init progress bar
            total_size = int(r.headers.get("content-length", 0))
            tqdm_args = {
                "total": total_size,
                "unit": "iB",
                "unit_scale": True,
            }
            with tqdm(**tqdm_args) as pbar:
                with open(to_path, "wb") as f:
                    for chunk in r.iter_content(chunk_size=1024):
                        f.write(chunk)
                        pbar.update(len(chunk))  # update progress bar
    except requests.exceptions.RequestException:
        logger.error(f"downloading of {from_url} failed")
        return
    else:
        return to_path


def decompress(compressed_path):
    """Decompress a bz2 file here. Overwrite previous version. Delete
    compressed file after decompression.
    """
    in_path = Path(compressed_path)
    out_path = in_path.parent.joinpath(in_path.stem)

    # rename latest decompressed file at this path (enables file comparison)
    indicate_as_old(out_path)

    try:
        with bz2.open(in_path) as in_f:
            with open(out_path, "wb") as out_f:
                data = in_f.read()
                out_f.write(data)
        in_path.unlink()
    except FileNotFoundError:
        logger.error(f"{compressed_path} not found by file decompressor")
        return
    except OSError:
        logger.error(f"{compressed_path} is not a valid compressed file")
        return
    else:
        return out_path


def extract(archive_path):
    """Extract here all files in an archive. Overwrite previous versions.
    Delete archive after extraction.
    """
    arx_path = Path(archive_path)

    try:
        with tarfile.open(arx_path) as tar:
            out_filenames = tar.getnames()
            out_paths = [arx_path.parent.joinpath(fn) for fn in out_filenames]
            # rename latest extracted files at this path for file comparison
            for fp in out_paths:
                indicate_as_old(fp)

            tar.extractall(arx_path.parent)

        arx_path.unlink()
    except FileNotFoundError:
        logger.error(f"{arx_path} not found by file extractor")
        return []
    except tarfile.ReadError:
        logger.error(f"{arx_path} is not an extractable archive")
        return []
    else:
        return out_paths


def fetch(from_url, to_directory):
    """Download a file, decompress it, extract it and delete temporary files.
    Overwrite previous versions.
    """
    logger.info(f"downloading {from_url}")
    dl_path = download(from_url, to_directory)

    if not dl_path:
        return []
    elif str(dl_path).endswith(".bz2"):
        logger.info(f"decompressing {dl_path.name}")
        uz_path = decompress(dl_path)
        if str(uz_path).endswith(".tar"):
            logger.info(f"extracting {uz_path.name}")
            out_paths = extract(uz_path)
            return out_paths
        else:
            return [uz_path]
    else:
        return [dl_path]


class lazy_property(object):
    """Implements a lazy property decorator. Lazy attributes are computed
    attributes that are evaluated only once, the first time they are used.
    Subsequent uses return the results of the first call.
    """

    def __init__(self, f):
        self.f = f
        self.name = f.__name__

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        val = self.f(instance)
        setattr(instance, self.name, val)

        return val


def get_filestem(url):
    """Get the stem of the file at this url."""
    return url.rsplit("/", 1)[-1].split(".", 1)[0]


def get_endpoint(url):
    """Get the parent url of this url."""
    return url.rsplit("/", 1)[0]


def get_byte_size_of_row(row, delimiter):
    """Get the byte size of this row split by this delimiter."""
    line = delimiter.join(row) + "\n"

    return len(line.encode("utf-8"))


def get_extended_name(file_path, extension):
    """Appends the name of this file"""
    return f"{file_path.stem}_{extension}{file_path.suffix}"


def indicate_as_old(file_path):
    """Show that a file is an old version by including 'old' in its name
    'foobar.txt' -> 'foobar_old.txt'
    """
    if file_path.is_file():
        name_of_old = get_extended_name(file_path, "old")
        file_path.replace(file_path.with_name(name_of_old))


def count_csv_columns(csv_path, delimiter):
    """Count the columns in a CSV file"""
    nb_cols = None
    try:
        with open(csv_path) as f:
            rd = csv.reader(f, delimiter=delimiter)
            nb_cols = len(next(rd))
    except FileNotFoundError:
        logger.exception("csv file not found")
    finally:
        return nb_cols



import logging

from .config import DATA_DIR
from .datafile import DataFile
from .exceptions import NoDataFile
from .utils import get_extended_name, lazy_property
from .version import version

logger = logging.getLogger(__name__)


class Tags:
    """The tags associated with each sentence."""

    _table = "tags"
    _dir = DATA_DIR.joinpath(_table)

    def __init__(self, language, scope="all"):

        # the language code of the sentences (ISO-639 code most of the time)
        self._lg = language
        # rows that are iterated through
        self._sp = scope

    def __iter__(self):

        if self._sp == "all":
            fpath = self.path
        else:
            fname = get_extended_name(self.path, self._sp)
            fpath = Tags._dir.joinpath(fname)

        try:
            fieldnames = ["sentence_id", "tag_name"]

            for row in DataFile(fpath, delimiter="\t"):
                row = {fieldnames[i]: x for i, x in enumerate(row)}

                yield Tag(**row)

        except NoDataFile:
            msg = (
                f"no '{self._sp}' data locally available for the "
                f"'{Tags._table}' table in {self._lg}."
            )

            logger.warning(msg)

    @property
    def language(self):
        """Get the language of the sentences tagged."""
        return self._lg

    @property
    def stem(self):
        """Get the stem of the file where the tags for this language
        pair are saved.
        """
        return f"{self._lg}_{Tags._table}"

    @property
    def filename(self):
        """Get the name of the file of these tagged sentences."""
        return f"{self.stem}.tsv"

    @property
    def path(self):
        """Get the path of the tagged sentences' datafile."""
        return Tags._dir.joinpath(self.filename)

    @lazy_property
    def get_version(self):
        """Get the version of the downloaded data of these tagged sentences."""
        return version[self.stem]


class Tag:
    """A tag associated to a sentence from the Tatoeba corpus."""

    def __init__(
        self,
        sentence_id,
        tag_name,
    ):

        self._id = sentence_id
        self._tag = tag_name

    @property
    def sentence_id(self):
        """Get the id of the sentence tagged."""
        return int(self._id)

    @property
    def tag_name(self):
        """Get the name of the tag."""
        return self._tag



import csv
import logging
from pathlib import Path

from tqdm import tqdm

from .buffer import Buffer
from .difference import compare_csv
from .exceptions import NoDataFile
from .utils import get_byte_size_of_row, lazy_property, get_extended_name
from .version import version

logger = logging.getLogger(__name__)


class DataFile:
    """A file containing table data."""

    def __init__(
        self, file_path, delimiter="\t", quoting=csv.QUOTE_NONE, text_col=-1
    ):
        # the local path of this data file
        self._fp = Path(file_path)
        # the delimiter that distinguishes table columns
        self._dm = delimiter
        # wether the fields are quoted or not
        self._qt = quoting
        # the column that must not be split by delimiters
        self._tc = text_col

    def __iter__(self):

        try:
            with open(self.path, encoding="utf-8") as f:
                for row in _custom_reader(f, self._dm, self._tc):
                    yield row
        except FileNotFoundError:
            logger.debug(f"{self.path} datafile not found")
            raise NoDataFile
        except RuntimeError:  # empty file
            pass

    def find_changes(self, index_col_keys=None):
        """Compare this file with its older version if there is one"""
        logger.info(f"finding changes in {self.name}")

        path_old = self._get_side_path("old")
        diffs = compare_csv(
            path_old,
            self._fp,
            delimiter=self._dm,
            index_col_keys=index_col_keys,
        )
        for tag, df in diffs.items():
            out_path = self._get_side_path(tag)
            df.to_csv(
                path_or_buf=out_path,
                sep=self._dm,
                header=False,
                index=False,
                quoting=self._qt,
            )

    def index(self, key_column, value_column):
        """Maps values from two columns of this datafile."""
        try:
            ind = {
                row[key_column]: row[value_column]
                for row in self
                if len(row) > key_column and len(row) > value_column
            }
        except NoDataFile:
            ind = {}

        return ind

    def split(self, columns=[], index=None):
        """Split the file according to the values mapped by the index
        in a chosen set of columns.
        """
        logger.info(f"splitting {self.name}")

        # init the progress bar
        pbar = tqdm(total=self.size, unit="iB", unit_scale=True)
        # init buffer
        buffer = Buffer(self.path.parent, delimiter=self.delimiter)
        # classify the rows
        for row in self:
            mapped_fields = _get_mapped_fields(row, columns, index)

            if all(mapped_fields):
                fname = self._get_out_filename(mapped_fields)
                buffer.add(row, fname)

            # imcrement progress bar by the byte size of the row
            pbar.update(get_byte_size_of_row(row, self.delimiter))

        buffer.clear()

        pbar.close()

    def _get_out_filename(self, mapped_fields):
        """Get the name of the file that corespond to this mapped fields."""
        mapped_fields_string = "-".join(mapped_fields)
        ext = "tsv" if self._dm == "\t" else "csv"

        return f"{mapped_fields_string}_{self.stem}.{ext}"

    def _get_side_path(self, side_tag):
        """Get the path of a side datafile"""
        extended_name = get_extended_name(self._fp, side_tag)

        return self._fp.parent.joinpath(extended_name)

    @property
    def path(self):
        """Get the path of this datafile"""
        return self._fp

    @property
    def delimiter(self):
        """Get the string that delimitate fields in this datafile"""
        return self._dm

    @property
    def name(self):
        """Get the name of this datafile"""
        return self._fp.name

    @property
    def stem(self):
        """Get the name of this datafile without its extension"""
        return self._fp.stem

    @property
    def size(self):
        """Get the byte size of this data file"""
        if self._fp.is_file():
            return self._fp.stat().st_size
        else:
            return 0

    @lazy_property
    def version(self):
        """Get the version datetime of this datafile"""
        return version[self.stem]


def _unsplit_field(row, nb_cols, delimiter, index_field):
    """Regroup the chosen field of a csv row if split by mistake. Useful if
    the fields are not quoted or if extra delimiters are not escaped.
    """
    if index_field < 0:
        index_field = nb_cols + index_field

    nb_extra = len(row) - nb_cols
    if nb_extra > 0:
        fields_to_join = row[index_field : index_field + nb_extra + 1]
        row[index_field] = delimiter.join(fields_to_join)
        del row[index_field + 1 : index_field + nb_extra + 1]

    return row


def _custom_reader(string_io, delimiter, text_col):
    """A customized version of csv.reader that:
    - unsplit unquoted field that contain delimiters
    - regroup unquoted multiline fields (i.e. containing newline characters)
    """
    reader = csv.reader(
        string_io,
        delimiter=delimiter,
        quoting=csv.QUOTE_NONE,
    )
    # count the number of columns in a regular row
    nb_cols = len(next(reader))
    string_io.seek(0)

    real_row = []
    for row in reader:
        # regroup text field if split by delimiter
        if text_col:
            row = _unsplit_field(row, nb_cols, delimiter, text_col)

        # regroup multiline end fields
        if len(row) == nb_cols:
            if real_row:
                yield real_row
                real_row = []
            real_row.extend(row)
        elif len(row) == 1 and real_row:
            real_row[-1] += " " + row[0]
        elif not row and real_row:
            real_row[-1] += " "
        else:
            logger.debug(f"row skipped: {row}")
    if real_row:
        yield real_row


def _get_mapped_fields(row, columns, index):
    """For this row, get the values of the fields in these columns.
    If index is True, get the value mapped by the index dictionary.
    """
    mapped_fields = []
    for col in columns:
        if len(row) > col:
            val = row[col]

            if index:
                val = index.get(val)

            mapped_fields.append(val)
        else:
            mapped_fields.append("")
            logger.debug(f"{row} does not have column {col}")

    return mapped_fields



from .config import DATA_DIR
from .utils import fetch, get_filestem
from .version import version


class Download:
    """A file download."""

    def __init__(self, url, version):
        # the url from which the file is downloaded
        self._url = url
        # the datetime used as the version of the file
        self._vs = version

    def fetch(self):
        """Download, decompress, extract, delete tamporary files, update
        local version value.
        """
        fetched = fetch(self.from_url, self.out_dir)
        if fetched:
            version[self.stem] = self.version

        return fetched

    @property
    def from_url(self):
        """Get the url from which the file is fetched."""
        return self._url

    @property
    def version(self):
        """Get the version of the downloaded file."""
        return self._vs

    @property
    def stem(self):
        """Get the stem of the downloaded file
        'https://foo.bar/foobar.txt' -> 'foobar'
        """
        return get_filestem(self.from_url)

    @property
    def table(self):
        """Get the name of the table from which this datafile is extracted."""
        for tbl in (
            "sentences_base",
            "sentences_detailed",
            "sentences_CC0",
            "transcriptions",
            "links",
            "tags",
            "sentences_in_lists",
            "sentences_with_audio",
            "user_languages",
        ):
            if self.stem.endswith(tbl):
                return tbl

        for tbl in (
            "user_lists",
            "jpn_indices",
            "queries",
        ):
            if self.stem == tbl:
                return tbl

        return

    @property
    def out_dir(self):
        """Get the path of the directory into which the downloaded file is
        saved.
        """
        return DATA_DIR.joinpath(self.table)



import logging

from .config import DIFFERENCE_TABLES, INDEX_SPLIT_TABLES, SIMPLE_SPLIT_TABLES
from .datafile import DataFile
from .download import Download
from .exceptions import NotAvailableLanguage, NotAvailableTable
from .jpn_indices import JpnIndices
from .links import Links
from .queries import Queries
from .sentences_base import SentencesBase
from .sentences_cc0 import SentencesCC0
from .sentences_detailed import SentencesDetailed
from .sentences_in_lists import SentencesInLists
from .sentences_with_audio import SentencesWithAudio
from .table import Table
from .tags import Tags
from .transcriptions import Transcriptions
from .update import check_languages, check_tables, check_updates
from .user_languages import UserLanguages
from .user_lists import UserLists
from .utils import lazy_property

logger = logging.getLogger(__name__)


class Tatoeba:
    """A handler for interacting with the local Tatoeba's data
    It enables users to download and parse Tatoeba export files
    """

    def sentences_detailed(
        self, language, scope="all", update=True, verbose=False
    ):
        """Iterates through all sentences in this language.

        Parameters
        ----------
        language : str
            The IS0 639-3 code of a Tatoeba supported language.
            Call the 'all_languages' attribute to get the list
            of all supported languages
        scope : str
            The scope of the iteration is "all", "added" or "removed"
        update : bool
            The data file is updated before being read when update is
            activated
        verbose : bool
            Update steps are printed when verbose is set to True

        Returns
        -------
        iterator
            SentenceDetailed instances with sentence_id, lang,
            text, username, date_added and date_last_modified
            attributes.
        """
        if update:
            self.update(["sentences_detailed"], [language], verbose=verbose)

        return SentencesDetailed(language=language, scope=scope).__iter__()

    def sentences_base(
        self, language, scope="all", update=True, verbose=False
    ):
        """Iterates through all sentences' bases in this language.

        Parameters
        ----------
        language : str
            The IS0 639-3 code of a Tatoeba supported language.
            Call the 'all_languages' attribute to get the list
            of all supported languages
        scope : str
            The scope of the iteration is "all", "added" or "removed"
        update : bool
            The data file is updated before being read when update is
            activated
        verbose : bool
            Update steps are printed when verbose is set to True

        Returns
        -------
        iterator
            SentenceBase instances with sentence_id and
            base_of_the_sentence attributes
        """
        if update:
            self.update(["sentences_base"], [language], verbose=verbose)

        return SentencesBase(language=language).__iter__()

    def sentences_CC0(self, language, scope="all", update=True, verbose=False):
        """Iterate through all sentences in this language with a CC0
        license

        Parameters
        ----------
        language : str
            The IS0 639-3 code of a Tatoeba supported language.
            Call the 'all_languages' attribute to get the list
            of all supported languages.
        scope : str
            The scope of the iteration is "all", "added" or "removed"
        update : bool
            The data file is updated before being read when update is
            activated
        verbose : bool
            Update steps are printed when verbose is set to True

        Returns
        -------
        iterator
            SentenceCC0 instances with sentence_id, lang, text and
            date_last_modified attributes.
        """
        if update:
            self.update(["sentences_CC0"], [language], verbose=verbose)

        return SentencesCC0(language=language).__iter__()

    def links(
        self,
        source_language,
        target_language,
        scope="all",
        update=True,
        verbose=False,
    ):
        """Iterates through all links between sentences in this source
        language and sentences in this target language

        Parameters
        ----------
        source_language : str
            The IS0 639-3 code of a Tatoeba supported language.
            Call the 'all_languages' attribute to get the list
            of all supported languages.
        target_language : str
            The IS0 639-3 code of a Tatoeba supported language.
            Call the 'all_languages' attribute to get the list
            of all supported languages.
        scope : str
            The scope of the iteration is "all", "added" or "removed"
        update : bool
            The data file is updated before being read when update is
            activated
        verbose : bool
            Update steps are printed when verbose is set to True

        Returns
        -------
        iterator
            Link instances with sentence_id and translation_id
            attributes
        """
        if update:
            self.update(
                ["links"],
                [source_language, target_language],
                oriented_pair=True,
                verbose=verbose,
            )

        return Links(
            source_language=source_language, target_language=target_language
        ).__iter__()

    def tags(self, language, scope="all", update=True, verbose=False):
        """Iterates through all tagged sentences in this language.

        Parameters
        ----------
        language : str
            The IS0 639-3 code of a Tatoeba supported language.
            Call the 'all_languages' attribute to get the list
            of all supported languages.
        scope : str
            The scope of the iteration is "all", "added" or "removed"
        update : bool
            The data file is updated before being read when update is
            activated
        verbose : bool
            Update steps are printed when verbose is set to True

        Returns
        -------
        iterator
            Tag instances with sentence_id and tag_name attributes
        """
        if update:
            self.update(["tags"], [language], verbose=verbose)

        return Tags(language=language).__iter__()

    def user_lists(self, scope="all", update=True, verbose=False):
        """Iterate trough all sentences' lists

        Parameters
        ----------
        scope : str
            The scope of the iteration is "all", "added" or "removed"
        update : bool
            The data file is updated before being read when update is
            activated
        verbose : bool
            Update steps are printed when verbose is set to True

        Returns
        -------
        iterator
            UserList instances with list_id, username, date_created,
            date_last_modified, list_name and editable_by attributes
        """
        if update:
            self.update(["user_lists"], [], verbose=verbose)

        return UserLists().__iter__()

    def sentences_in_lists(
        self, language, scope="all", update=True, verbose=False
    ):
        """Iterates through all sentences in this language which
        are in a list

        Parameters
        ----------
        language : str
            The IS0 639-3 code of a Tatoeba supported language.
            Call the 'all_languages' attribute to get the list
            of all supported languages.
        scope : str
            The scope of the iteration is "all", "added" or "removed"
        update : bool
            The data file is updated before being read when update is
            activated
        verbose : bool
            Update steps are printed when verbose is set to True

        Returns
        -------
        iterator
            SentenceInList instances with list_id and sentence_id
            attributes
        """
        if update:
            self.update(["sentences_in_lists"], [language], verbose=verbose)

        return SentencesInLists(language=language).__iter__()

    def jpn_indices(self, scope="all", update=True, verbose=False):
        """Iterates through all Japanese indices

        Parameters
        ----------
        scope : str
            The scope of the iteration is "all", "added" or "removed"
        update : bool
            The data file is updated before being read when update is
            activated
        verbose : bool
            Update steps are printed when verbose is set to True

        Returns
        -------
        iterator
            JpnIndex instances with sentence_id, meaning_id and text
            attributes
        """
        if update:
            self.update(["jpn_indices"], [], verbose=verbose)

        return JpnIndices().__iter__()

    def sentences_with_audio(
        self, language, scope="all", update=True, verbose=False
    ):
        """Iterates through sentences with audio file

        Parameters
        ----------
        language : str
            The IS0 639-3 code of a Tatoeba supported language.
            Call the 'all_languages' attribute to get the list
            of all supported languages
        scope : str
            The scope of the iteration is "all", "added" or "removed"
        update : bool
            The data file is updated before being read when update is
            activated
        verbose : bool
            Update steps are printed when verbose is set to True

        Returns
        -------
        iterator
            SentenceWithAudio instances with sentence_id, username,
            license and attribution_url attributes
        """
        if update:
            self.update(["sentences_with_audio"], [language], verbose=verbose)

        return SentencesWithAudio(language=language).__iter__()

    def user_languages(
        self, language, scope="all", update=True, verbose=False
    ):
        """Iterates through all users' skills in this language

        Parameters
        ----------
        language : str
            The IS0 639-3 code of a Tatoeba supported language.
            Call the 'all_languages' attribute to get the list
            of all supported languages
        scope : str
            The scope of the iteration is "all", "added" or "removed"
        update : bool
            The data file is updated before being read when update is
            activated
        verbose : bool
            Update steps are printed when verbose is set to True

        Returns
        -------
        iterator
            UserLanguage instances with lang, skill_level,
            username and details attributes
        """
        if update:
            self.update(["user_languages"], [language], verbose=verbose)

        return UserLanguages(language=language).__iter__()

    def transcriptions(
        self, language, scope="all", update=True, verbose=False
    ):
        """Iterate through all transcriptions for this language

        Parameters
        ----------
        language : str
            The IS0 639-3 code of a Tatoeba supported language.
            Call the 'all_languages' attribute to get the list
            of all supported languages
        scope : str
            The scope of the iteration is "all", "added" or "removed"
        update : bool
            The data file is updated before being read when update is
            activated
        verbose : bool
            Update steps are printed when verbose is set to True

        Returns
        -------
        iterator
            Transcription instances with sentence_id, lang,
            script_name, username and transcription attributes
        """
        if update:
            self.update(["transcriptions"], [language], verbose=verbose)

        return Transcriptions(language=language).__iter__()

    def queries(self, language, scope="all", update=True, verbose=False):
        """Iterate through all queries for this language

        Parameters
        ----------
        language : str
            The IS0 639-3 code of a Tatoeba supported language.
            Call the 'all_languages' attribute to get the list
            of all supported languages
        scope : str
            The scope of the iteration is "all", "added" or "removed"
        update : bool
            The data file is updated before being read when update is
            activated
        verbose : bool
            Update steps are printed when verbose is set to True

        Returns
        -------
        iterator
            Queries instances with date, language and content
            attributes
        """
        if update:
            self.update(["queries"], [language], verbose=verbose)

        return Queries(language=language, scope=scope).__iter__()

    def update(
        self, table_names, language_codes, oriented_pair=False, verbose=False
    ):
        """Updates the local Tatoeba datafiles:
        - download newest versions of the datafiles
        - split them by language (when nomonolingual versions not avialable)
        - identify the differences with the previously downloaded versions

        Parameters
        ----------
        table_names : list
            The names of the tables to update. Call the 'all_tables'
            attribute to get a list of all available tables
        language_codes : list
            ISO 639-3 codes of the languages for which local data is
            updated. Call the 'all_languages' attribute to get the list
            of all supported languages.
        oriented_pair : bool, optional
            Requires a pair of language codes where the first language
            is considered as source and the second as target,
            by default False
        verbose : bool, optional
            Verbosity of the logging, by default True

        Raises
        ------
        NotAvailableTable
            Indicates that at least one of the table naames pased as
            argument is not a valid Tatoeba table name.
        NotAvailableLanguage
            Indicates that at least one of the language code pased as
            argument is not supported by Tatoeba.
        """
        if not table_names and not language_codes:
            return

        # check if the tables are available
        not_available_tables = set(table_names) - set(self.all_tables)
        if not_available_tables:
            raise NotAvailableTable(not_available_tables)

        # check if the language codes are available
        not_available_langs = set(language_codes) - set(self.all_languages)
        if not_available_langs:
            raise NotAvailableLanguage(not_available_langs)

        # sentences table added when required for splitting other files
        if (
            any(tn in INDEX_SPLIT_TABLES for tn in table_names)
            and "sentences_detailed" not in table_names
        ):
            table_names.append("sentences_detailed")

        # get the urls of the datafiles that need an update
        to_download = check_updates(
            table_names,
            language_codes,
            oriented_pair=oriented_pair,
            verbose=verbose,
        )

        # download the files of the update
        download_paths = []
        for url, vs in to_download.items():
            download_paths.extend(Download(url, vs).fetch())

        language_index = {}
        monolang_datafiles = []
        for fp in download_paths:
            # split the multilingual datafiles by language
            if fp.stem in table_names:
                table = Table(fp.stem, language_codes)

                if table.name in INDEX_SPLIT_TABLES:
                    if not language_index:
                        logger.info("mapping sentence ids to languages")

                        sentence_table = Table(
                            "sentences_detailed", language_codes
                        )
                        language_index = sentence_table.index(0, 1)

                    monolang_datafiles.extend(table.classify(language_index))

                elif table.name in SIMPLE_SPLIT_TABLES:
                    monolang_datafiles.extend(table.classify())

            else:
                delimiter = "\t" if fp.suffix == ".tsv" else ","
                monolang_datafiles.append(DataFile(fp, delimiter=delimiter))

        # compare monolingual datafiles with their older version
        for df in monolang_datafiles:
            if any(tbl in df.stem for tbl in DIFFERENCE_TABLES):
                df.find_changes(index_col_keys=None)

        if verbose:
            updated = {fp.stem for fp in download_paths}
            updated |= {df.stem for df in monolang_datafiles}
            updated = sorted(list(updated))
            if updated:
                msg = "updated files: {}".format(", ".join(updated))
                logger.info(msg)

    @property
    def all_tables(self):
        """All tables that are downloadable from tatoeba.org

        Returns
        -------
        list
            See https://tatoeba.org/eng/downloads for more information.
        """
        return check_tables()

    @lazy_property
    def all_languages(self):
        """All languages with at least one sentence on tatoeba.org

        Returns
        -------
        list
            See https://tatoeba.org/eng/stats/sentences_by_language
            for more information.
        """
        return check_languages()



import csv
import logging
from pathlib import Path
from sys import getsizeof

logger = logging.getLogger(__name__)


class Buffer:
    """A buffer temporarily stores data and then appends it into out files
    when full. It is useful to avoid memory overflow when handling very large
    data files.
    """

    def __init__(self, out_dir, delimiter="\t", max_size=10000):
        # directory path where out files are saved.
        self._dir = Path(out_dir)
        self._dir.mkdir(parents=True, exist_ok=True)
        self._fps = set(self._dir.iterdir())  # files already in the out_dir
        # the feed delimiter used in the out files
        self._dm = delimiter
        # maximum number elements in a buffer
        self._max = max_size
        # the buffer data is classified in a dict. The dict keys are named
        # after the out filenames the data is directed to.
        self._data = {}

    def __getitem__(self, out_fname):

        return self._data[out_fname]

    def __setitem__(self, out_fname, data):

        self._data[out_fname] = data

    def __len__(self):

        return len(self._data)

    def add(self, elt, out_fname):
        """Adds an element into the buffer linked to 'out_fname'. Once the
        buffer is full, this element is appended to the file at
        'out_dir/out_fname'.
        """
        if out_fname not in self._data:
            self._data[out_fname] = []
            # reinitialize the out file
            out_fp = Path(self._dir, out_fname)
            if out_fp in self._fps:
                out_fp.unlink()
                self._fps.remove(out_fp)

        self._data.setdefault(out_fname, []).append(elt)

        if getsizeof(self._data[out_fname]) > self._max:
            self._save(out_fname)

    def _save(self, out_fname, end=False):
        """Appends buffered elements into their out datafile and then clears
        the buffer.
        """
        data = list(self._data[out_fname])
        out_fp = Path(self._dir, f"{out_fname}.part")
        try:
            with open(out_fp, mode="a") as f:
                wt = csv.writer(f, delimiter=self._dm)
                wt.writerows(data)
        except FileNotFoundError:
            logger.debug(f"an error occured when opening {out_fp}")
        else:
            self._data[out_fname].clear()

            if end:
                # removes '.part' extension
                out_fp.rename(out_fp.parent.joinpath(out_fp.stem))

    def clear(self):
        """Saves the data remaining in the buffer into the corresponding
        outfiles.
        """
        for out_fname in self._data.keys():
            self._save(out_fname, end=True)

        self._data.clear()

    @property
    def directory(self):
        """Get the path of the directory where the buffer files are saved."""
        return self._dir

    @property
    def out_filenames(self):
        """List every out file name."""
        return list(self._data.keys())



import logging

from .config import DATA_DIR
from .datafile import DataFile
from .exceptions import NoDataFile
from .utils import get_extended_name, lazy_property
from .version import version

logger = logging.getLogger(__name__)


class SentencesInLists:
    """The Tatoeba sentences that are contained in a list."""

    _table = "sentences_in_lists"
    _dir = DATA_DIR.joinpath(_table)

    def __init__(self, language, scope="all"):

        # the language code of the sentences (ISO-639 code most of the time)
        self._lg = language
        # rows that are iterated through
        self._sp = scope

    def __iter__(self):

        if self._sp == "all":
            fpath = self.path
        else:
            fname = get_extended_name(self.path, self._sp)
            fpath = SentencesInLists._dir.joinpath(fname)

        fieldnames = [
            "list_id",
            "sentence_id",
        ]

        try:
            for row in DataFile(fpath, delimiter="\t"):
                row = {fieldnames[i]: x for i, x in enumerate(row)}

                yield SentenceInList(**row)

        except NoDataFile:
            msg = (
                f"no '{self._sp}' data locally available for the "
                f"'{SentencesInLists._table}' table in {self._lg}."
            )

            logger.warning(msg)

    @property
    def language(self):
        """Get the language of the sentences."""
        return self._lg

    @property
    def stem(self):
        """Get the stem of the name of the datafile"""
        return f"{self._lg}_{SentencesInLists._table}"

    @property
    def filename(self):
        """Get the name of the datafile"""
        return f"{self.stem}.tsv"

    @property
    def path(self):
        """Get the path of the datafile."""
        return SentencesInLists._dir.joinpath(self.filename)

    @lazy_property
    def version(self):
        """Get the version of the downloaded data of these sentences"""
        return version[self.stem]


class SentenceInList:
    """A sentence from the Tatoeba corpus which is in a list."""

    def __init__(self, list_id, sentence_id):

        self._lid = list_id
        self._sid = sentence_id

    @property
    def list_id(self):
        """Get the id of the list."""
        return int(self._lid)

    @property
    def sentence_id(self):
        """Get the id of the sentence."""
        return int(self._sid)



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tatoebatools",
    version="0.0.7",
    author="L.Beaudoux",
    description="A library for downloading and reading data from Tatoeba",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/l-bdx/tatoebatools",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
    ],
    python_requires=">=3.6",
    install_requires=[
        "beautifulsoup4>=4.9.0",
        "pandas>=1.1.1",
        "requests>=2.23.0",
        "tqdm>=4.46.0",
    ],
)



class NoDataFile(Exception):
    """Raised when the datafile read is not locally available"""


class NotAvailableTable(Exception):
    """Raised when a table passed as argument to the update is not
    available"""

    def __init__(self, not_available_tables):

        s1 = len(not_available_tables)
        s2 = "s" if s1 > 1 else ""
        s3 = ", ".join(not_available_tables)
        msg = f"{s1} not available table{s2}: {s3}"

        super().__init__(msg)

        self.table_names = not_available_tables


class NotAvailableLanguage(Exception):
    """Raised when a language passed as argument to the update is not
    available"""

    def __init__(self, not_available_langs):

        s1 = len(not_available_langs)
        s2 = "s" if s1 > 1 else ""
        s3 = ", ".join(not_available_langs)
        msg = f"{s1} not available language{s2}: {s3}"

        super().__init__(msg)

        self.language_codes = not_available_langs


class NotLanguagePair(Exception):
    """Raised when oriented_pair is activated and the number of
    languages passed as argument to the update differs from two
    """

    def __init__(self, lang_codes):

        s = ", ".join(lang_codes)
        msg = f"activated 'oriented_pair' requires [{s}] to be a language pair"

        super().__init__(msg)

        self.language_codes = lang_codes
